from pyomo.environ import SolverFactory, Objective, ConcreteModel, Var, NonNegativeReals, Constraint, maximize, Set


def create_model(net_value):

    lenders = sorted([key for key, value in net_value.items() if value < 0])
    loaners = sorted([key for key, value in net_value.items() if value >= 0])

    model = ConcreteModel()

    model.Loaners = Set(initialize=loaners)
    model.Lenders = Set(initialize=lenders)

    model.Amount = Var(model.Loaners, model.Lenders, domain=NonNegativeReals)

    def obj_expression(model):
        return sum(model.Amount[i, j] * model.Amount[i, j]
                   for i in model.Loaners for j in model.Lenders)

    model.OBJ = Objective(rule=obj_expression, sense=maximize)

    def loaner_constraint_rule(model, i):
        return sum(model.Amount[i, j]
                   for j in model.Lenders) <= net_value[i]

    def lender_constraint_rule(model, j):
        return sum(model.Amount[i, j]
                   for i in model.Loaners) <= -net_value[j]

    model.AmountConstraintLoaner = Constraint(model.Loaners,
                                              rule=loaner_constraint_rule)
    model.AmountConstraintLender = Constraint(model.Lenders,
                                              rule=lender_constraint_rule)

    return model


def solve(model):

    opt = SolverFactory('ipopt')
    instance = model
    results = opt.solve(instance)
    instance.solutions.store_to(results)

    return results



__all__ = ("__title__",
           "__summary__",
           "__uri__",
           "__version__",
           "__author__",
           "__email__",
           "__license__",
           "__copyright__", )


def __get_version():
    from os import path
    f = path.abspath(path.dirname(__file__))
    return (open(path.join(f, 'version.py')).read())


exec(__get_version())

__title__ = "py-iou"
__summary__ = "A python package to find optimal number of transactions betweeen friends"
__uri__ = "https://github.com/kdheepak/iou"

__author__ = "Dheepak Krishnamurthy"
__email__ = "kdheepak89@gmail.com"

__license__ = "Revised BSD License"
__copyright__ = "Copyright 2016 Dheepak Krishnamurthy"



#!/usr/bin/env python
"""Run module."""

import click
from . import data
from . import version
from . import ReferenceModel

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--data', type=click.Path(), help='Path to transactions.csv')
@click.option('--verbose', default=False, is_flag=True, help='Verbose mode')
@click.option('--color', default='green', help='Color of text')
@click.option('--verbose-color', default='white', help='Color of text')
@click.version_option(version.__version__, '-v', '--version')
def main(**kwargs):
    """iou - Find optimal transactions."""

    color = kwargs['color']

    try:
        if not any([kwargs[key] for key in ['data']]):
            help_str = "{}".format(click.get_current_context().get_help())
            click.secho(help_str)
            click.get_current_context().exit()

        def show_item(item):
            if item is not None:
                return item

        click.echo("")
        click.secho("Find the Optimal Number of Transactions:", fg=color, bold=True)
        click.echo("")

        with click.progressbar(
            ('Getting data', 'Creating model', 'Solving', 'Finished'),
                label='iou:',
                item_show_func=show_item) as bar:
            for item in bar:
                if item == 'Getting data':
                    path_to_file = kwargs['data']
                    array, people = data.get_transactions_from_file(path_to_file)
                    net_value = data.compress_edges(data.create_graph(array, people))
                elif item == 'Creating model':
                    model = ReferenceModel.create_model(net_value)
                elif item == 'Solving':
                    ReferenceModel.solve(model)

        if kwargs['verbose']:
            color = kwargs['verbose_color']
            click.echo("")
            for row_number, row in enumerate(array):
                number_of_contributors = sum([float(item) for item in row[3:]])

                click.secho("{number_of_contributors}".format(number_of_contributors=number_of_contributors), bold=True, fg=color, nl=False)
                click.secho(" people owe ", nl=False)
                click.secho("{person}".format(person=row[2]), bold=True, fg=color, nl=False)
                click.secho(" ", nl=False)
                click.secho("{amount}".format(amount=row[1]), bold=True, fg=color, nl=False)
                click.secho(" for ", nl=False)
                click.secho("{expense}".format(expense=row[0]), bold=True, fg=color)

            click.echo("")
            for person, value in net_value.items():
                if value > 0:
                    click.secho("{person}".format(
                        person=person), bold=True, fg=color, nl=False)
                    click.secho("'s net spend =", nl=False)
                    click.secho(" {value}".format(value=value), bold=True, fg=color)

            click.echo("")
            for person, value in net_value.items():
                if value < 0:
                    click.secho("{person}".format(
                        person=person), bold=True, fg=color, nl=False)
                    click.secho("'s net spend =", nl=False)
                    click.secho(" {value}".format(value=value), bold=True, fg=color)

        click.echo("")

        color = kwargs['color']

        click.secho("iou results:")

        click.echo("")

        for i in model.Amount:
            if model.Amount[i].value == 0:
                pass
            else:
                click.secho("{}".format(i[0]), fg=color, bold=True, nl=False)
                click.secho(" owes", nl=False)
                click.secho(" {}".format(i[1]), fg=color, bold=True, nl=False)
                click.secho(" a total of ", nl=False)
                click.secho("${0:.2f}".format(model.Amount[i].value), fg=color, bold=True)

        click.echo("")

    except Exception as e:
        click.echo('')
        raise click.ClickException("{}\n\nCheck the help (--help) on how to use iou or contact the developer.".format(
            e.message))


if __name__ == '__main__':
    main()



from setuptools import setup
try:
    import iou
except ImportError:
    iou = object()
    iou.__title__ = "py-iou"
    iou.__summary__ = "A python package to find optimal number of transactions betweeen friends"
    iou.__uri__ = "https://github.com/kdheepak/iou"

    iou.__author__ = "Dheepak Krishnamurthy"
    iou.__email__ = "kdheepak89@gmail.com"

    iou.__license__ = "Revised BSD License"
    iou.__copyright__ = "Copyright 2016 Dheepak Krishnamurthy"

from os import path
from pip.req import parse_requirements

here = path.abspath(path.dirname(__file__))

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except:
    with open(path.join(here, 'README.md')) as f:
        long_description = f.read()

install_requires = [str(ir.req)
                    for ir in parse_requirements(
                        path.join(here, 'requirements.txt'),
                        session=False)]

setup(
    name=iou.__title__,
    version=iou.__version__,
    description=iou.__summary__,
    long_description=long_description,
    license=iou.__license__,
    url=iou.__uri__,
    author=iou.__author__,
    author_email=iou.__email__,
    packages=["iou", "iou.data"],
    entry_points={
        "console_scripts": [
            "iou = iou.run:main",
        ],
    },
    install_requires=install_requires,

    # dependency_links=[
    # "git+ssh://git@github.com/kdheepak89/click.git@7.0#egg=click-7.0"
    # ]
)



import networkx as nx
import pandas as pd

def get_transactions_from_file(f):
    df = pd.read_csv(f)

    array = df.values
    people = df.columns[3:].values

    return array, people

def create_graph(array, people):

    G = nx.MultiDiGraph()

    for person_name in people:
        if person_name is None:
            pass
        else:
            G.add_node(person_name)

    for row_number, row in enumerate(array):
        number_of_contributors = sum([float(item) for item in row[3:]])
        for i, item in enumerate(row[3:]):
            if item != '0':
                G.add_edge(people[i], row[2], owe=float(
                    item) * float(row[1].strip('$')) / number_of_contributors)
        # print("{number_of_contributors} people owe {person} {amount} for {expense}".format(
            # number_of_contributors=number_of_contributors,
            # person=row[2],
            # expense=row[0],
            # amount=row[1])
        # )

    return G

def compress_edges(G):
    # empty dictionary
    net_value = {}
    for node in G.nodes():
        net_value[node] = 0

    for edge in G.edges(data=True):
        net_value[edge[0]] -= edge[2]['owe']
        net_value[edge[1]] += edge[2]['owe']

    net_value = {key: -value for key, value in net_value.items()}

    return net_value


